package main

import (
	"flag"
	"io/fs"
	"log"
	"net/http"
	"net/url"
	"os"
	"os/signal"
	"path"
	"path/filepath"
	"strings"
	"syscall"

	"srs.epita.fr/fic-server/libfic"
	"srs.epita.fr/fic-server/settings"
)

var DashboardDir string
var TeamsDir string

type ResponseWriterPrefix struct {
	real   http.ResponseWriter
	prefix string
}

func (r ResponseWriterPrefix) Header() http.Header {
	return r.real.Header()
}

func (r ResponseWriterPrefix) WriteHeader(s int) {
	if v, exists := r.real.Header()["Location"]; exists {
		r.real.Header().Set("Location", r.prefix+v[0])
	}
	r.real.WriteHeader(s)
}

func (r ResponseWriterPrefix) Write(z []byte) (int, error) {
	return r.real.Write(z)
}

func StripPrefix(prefix string, h http.Handler) http.Handler {
	if prefix == "" {
		return h
	}
	return http.HandlerFunc(func(w http.ResponseWriter, r *http.Request) {
		if prefix != "/" && r.URL.Path == "/" {
			http.Redirect(w, r, prefix+"/", http.StatusFound)
		} else if p := strings.TrimPrefix(r.URL.Path, prefix); len(p) < len(r.URL.Path) {
			r2 := new(http.Request)
			*r2 = *r
			r2.URL = new(url.URL)
			*r2.URL = *r.URL
			r2.URL.Path = p
			h.ServeHTTP(ResponseWriterPrefix{w, prefix}, r2)
		} else {
			h.ServeHTTP(w, r)
		}
	})
}

func main() {
	var baseURL string
	// Read paremeters from environment
	if v, exists := os.LookupEnv("FIC_BASEURL"); exists {
		baseURL = v
	}

	// Read parameters from command line
	var bind = flag.String("bind", "127.0.0.1:8082", "Bind port/socket")
	htpasswd_file := flag.String("htpasswd", "", "Restrict access with password, Apache htpasswd format")
	restrict_ip := flag.String("restrict-to-ips", "", "Restrict access to IP listed in this JSON array")
	flag.StringVar(&baseURL, "baseurl", baseURL, "URL prepended to each URL")
	staticDir := flag.String("static", "./htdocs-dashboard/", "Directory containing static files")
	flag.StringVar(&fic.FilesDir, "files", fic.FilesDir, "Base directory where found challenges files, local part")
	flag.StringVar(&DashboardDir, "dashbord", "./DASHBOARD", "Base directory where save public JSON files")
	flag.StringVar(&TeamsDir, "teams", "./TEAMS", "Base directory where save teams JSON files")
	flag.StringVar(&settings.SettingsDir, "settings", "./SETTINGSDIST", "Base directory where load and save settings")
	var fwdr = flag.String("forwarder", "", "URL of another dashboard where send traffic to, except static assets")
	flag.BoolVar(&fwdPublicJson, "fwdpublicjson", fwdPublicJson, "Also forward public.json files to forwarder")
	flag.Parse()

	log.SetPrefix("[public] ")

	// Sanitize options
	var err error
	log.Println("Checking paths...")
	if staticDir != nil && *staticDir != "" {
		if sDir, err := filepath.Abs(*staticDir); err != nil {
			log.Fatal(err)
		} else {
			log.Println("Serving pages from", sDir)
			staticFS = http.Dir(sDir)
		}
	} else {
		sub, err := fs.Sub(assets, "static")
		if err != nil {
			log.Fatal("Unable to cd to static/ directory:", err)
		}
		log.Println("Serving pages from memory.")
		staticFS = http.FS(sub)
	}
	if fic.FilesDir, err = filepath.Abs(fic.FilesDir); err != nil {
		log.Fatal(err)
	}
	if settings.SettingsDir, err = filepath.Abs(settings.SettingsDir); err != nil {
		log.Fatal(err)
	}
	if baseURL != "/" {
		baseURL = path.Clean(baseURL)
	} else {
		baseURL = ""
	}
	if fwdr != nil && len(*fwdr) > 0 {
		forwarder = fwdr
	}

	// Prepare graceful shutdown
	interrupt := make(chan os.Signal, 1)
	signal.Notify(interrupt, os.Interrupt, syscall.SIGTERM)

	app := NewApp(htpasswd_file, restrict_ip, baseURL, *bind)
	go app.Start()

	// Wait shutdown signal
	<-interrupt

	log.Print("The service is shutting down...")
	app.Stop()
	log.Println("done")
}
