package main

import (
	"context"
	"flag"
	"fmt"
	"log"
	"net/http"
	"net/url"
	"os"
	"os/signal"
	"path"
	"path/filepath"
	"strings"
	"syscall"

	"srs.epita.fr/fic-server/libfic"
	"srs.epita.fr/fic-server/qa/api"
)

var StaticDir string

type ResponseWriterPrefix struct {
	real   http.ResponseWriter
	prefix string
}

func (r ResponseWriterPrefix) Header() http.Header {
	return r.real.Header()
}

func (r ResponseWriterPrefix) WriteHeader(s int) {
	if v, exists := r.real.Header()["Location"]; exists {
		r.real.Header().Set("Location", r.prefix+v[0])
	}
	r.real.WriteHeader(s)
}

func (r ResponseWriterPrefix) Write(z []byte) (int, error) {
	return r.real.Write(z)
}

func StripPrefix(prefix string, h http.Handler) http.Handler {
	if prefix == "" {
		return h
	}
	return http.HandlerFunc(func(w http.ResponseWriter, r *http.Request) {
		if prefix != "/" && r.URL.Path == "/" {
			http.Redirect(w, r, prefix+"/", http.StatusFound)
		} else if p := strings.TrimPrefix(r.URL.Path, prefix); len(p) < len(r.URL.Path) {
			r2 := new(http.Request)
			*r2 = *r
			r2.URL = new(url.URL)
			*r2.URL = *r.URL
			r2.URL.Path = p
			h.ServeHTTP(ResponseWriterPrefix{w, prefix}, r2)
		} else {
			h.ServeHTTP(w, r)
		}
	})
}

func main() {
	// Read paremeters from environment
	if v, exists := os.LookupEnv("FIC_BASEURL"); exists {
		BaseURL = v
	}

	// Read parameters from command line
	var bind = flag.String("bind", "127.0.0.1:8083", "Bind port/socket")
	var dsn = flag.String("dsn", fic.DSNGenerator(), "DSN to connect to the MySQL server")
	flag.StringVar(&BaseURL, "baseurl", BaseURL, "URL prepended to each URL")
	flag.StringVar(&StaticDir, "static", "./htdocs-qa/", "Directory containing static files")
	flag.StringVar(&api.TeamsDir, "teams", "./TEAMS", "Base directory where save teams JSON files")
	flag.StringVar(&api.Simulator, "simulator", "", "Auth string to simulate (for development only)")
	flag.Parse()

	log.SetPrefix("[qa] ")

	// Sanitize options
	var err error
	log.Println("Checking paths...")
	if StaticDir, err = filepath.Abs(StaticDir); err != nil {
		log.Fatal(err)
	}
	if BaseURL != "/" {
		BaseURL = path.Clean(BaseURL)
	} else {
		BaseURL = ""
	}
	if api.Simulator != "" {
		if _, err := os.Stat(path.Join(api.TeamsDir, api.Simulator)); os.IsNotExist(err) {
			log.Fatal(err)
		}
	}

	// Database connection
	log.Println("Opening database...")
	if err = fic.DBInit(*dsn); err != nil {
		log.Fatal("Cannot open the database: ", err)
	}
	defer fic.DBClose()

	// Prepare graceful shutdown
	interrupt := make(chan os.Signal, 1)
	signal.Notify(interrupt, os.Interrupt, syscall.SIGTERM)

	srv := &http.Server{
		Addr:    *bind,
		Handler: StripPrefix(BaseURL, api.Router()),
	}

	// Serve content
	go func() {
		log.Fatal(srv.ListenAndServe())
	}()
	log.Println(fmt.Sprintf("Ready, listening on %s", *bind))

	// Wait shutdown signal
	<-interrupt

	log.Print("The service is shutting down...")
	srv.Shutdown(context.Background())
	log.Println("done")
}
